use {
	std::{
		collections::HashMap,
		any::type_name,
		time::{
			SystemTime,
			UNIX_EPOCH,
		},
	},
	webhook::client::WebhookClient,
	urlencoding::decode as url_decode,
	uuid::Uuid,
	chrono::DateTime,
	serde_json::{
		Value as Json,
		Map as JsonMap,
		json,
	},
	rsa::{
		pkcs1::{
			EncodeRsaPublicKey,
			EncodeRsaPrivateKey,
		},
		RsaPrivateKey,
		RsaPublicKey,
	},
	jsonwebtoken as jwt,
	jwt::{
		Header,
		Algorithm,
		TokenData,
		Validation,
		EncodingKey,
		DecodingKey,
		encode as jwt_encode,
		decode as jwt_decode,
	},
	skytable::{
		query,
		error::Error as SkyError,
	},
	http_body_util::BodyExt,
	hyper::{
		Request,
		HeaderMap,
		body::Incoming,
		header::{
			HeaderValue,
		},
	},
	crate::{
		Res,
		Result,
		DBPool,
		REFRESH_LIFETIME,
		TOKEN_LIFETIME,
		html::*,
		Users,
		Sites,
		APIV0_LIFETIME,
	},
};


#[allow(dead_code)]
pub fn type_of<T>(_: T) -> &'static str {
	type_name::<T>()
}

pub fn uuid_v4() -> Uuid {
	Uuid::new_v4()
}

pub fn build_html(body: &str) -> String {
	format!("{}{}{}{}", HEADER_HTML, CSS3, body, FOOTER_HTML)
}

pub async fn init_tables(pool: DBPool) -> Res<(), SkyError> {
	let mut con = pool.get().await.unwrap();

	let _ = con.query_parse::<()>(&query!("CREATE SPACE IF NOT EXISTS bitauth")).await;
	let _ = con.query_parse::<()>(&query!(r#"
		CREATE MODEL IF NOT EXISTS bitauth.users(
			uuid: string,
			login: string,
			password: string,
			email: string,
			tokens: list {type: string}
		)
	"#)).await;
	let _ = con.query_parse::<()>(&query!(r#"
		CREATE MODEL IF NOT EXISTS bitauth.users_uuid(
			login: string,
			uuid: string
		)
	"#)).await;
	let _ = con.query_parse::<()>(&query!(r#"
		CREATE MODEL IF NOT EXISTS bitauth.sites(
			uuid: string,
			uid: string,
			domain: string,
			pkey: binary,
			skey: binary
		)
	"#)).await;
	let _ = con.query_parse::<()>(&query!(r#"
		CREATE MODEL IF NOT EXISTS bitauth.tokens(
			uuid: string,
			uid: string,
			sid: string,
			ref: string,
			refend: uint32
		)
	"#)).await;
	let _ = con.query_parse::<()>(&query!(r#"
		CREATE MODEL IF NOT EXISTS bitauth.v0(
			session: string,
			login: string,
			uuid: string,
			expire: uint32
		)
	"#)).await;

	let q = con.query_parse::<Sites>(&query!("SELECT * FROM bitauth.sites WHERE uuid = 0")).await;
	if q.is_err() {
		let (skey, pkey) = rsa_gen();
		let _ = con.query_parse::<()>(&query!(
			"INSERT INTO bitauth.sites { uuid: ?, uid: ?, domain: ?, pkey: ?, skey: ? }",
			"0", "0", "", pkey, skey
		)).await;
	};

	Ok(())
}

pub fn double_split(body: String, first: &str, second: &str) -> HashMap<String, String> {
	body.split(first)
		.filter_map(|c| {
			c.split_once(second)
				.map(|(l, r)| (
					l.trim().to_owned(),
					format!("{}", url_decode(r).expect("UTF-8")).trim().to_owned()
				))
		})
		.collect::<HashMap<String, String>>()
}

pub fn time() -> u32 {
	SystemTime::now()
		.duration_since(UNIX_EPOCH)
		.unwrap()
		.as_secs() as u32
}

pub fn time_mcs() -> u128 {
	SystemTime::now()
		.duration_since(UNIX_EPOCH)
		.unwrap()
		.as_micros()
}

pub async fn discord_wh_send(text: String) {
	let url: &str = "https://discord.com/api/webhooks/1228566682902204537/RXaWpZplEGzU88O8c4mD4qzXp1PDBrrp93nGvdijaY7mBXp27xc0EsThHUeU0431PQOZ";
	let client: WebhookClient = WebhookClient::new(url);
	client.send(|message| message
		.username("Bit.Auth")
//		.avatar_url("")
		.embed(|embed| embed
			.title("Error")
			.description(&text)
//			.footer("Footer", Some(String::from(IMAGE_URL)))
//			.image(IMAGE_URL)
//			.thumbnail(IMAGE_URL)
//			.author("bitheaven", Some(String::from(IMAGE_URL)), Some(String::from(IMAGE_URL)))
//			.field("name", "value", false)
		)
	).await.unwrap();
}

pub fn set_cookie(headers: &mut HeaderMap, key: &str, value: &str) {
	let time = DateTime::from_timestamp((time() + REFRESH_LIFETIME) as i64, 0)
		.expect("REASON")
		.to_rfc2822();
	let time = time.replace("+0000", "GMT");
	headers.append(
		hyper::header::SET_COOKIE,
		HeaderValue::from_str(format!("{}={}; HttpOnly; Expires={}", key, value, time).as_str())
			.unwrap()
	);
}

pub fn set_location(headers: &mut HeaderMap, location: &str) {
	headers.append(
		hyper::header::LOCATION,
		HeaderValue::from_str(format!("{}", location).as_str()).unwrap()
	);
}

pub async fn get_body_from_request(mut req: Request<Incoming>) -> Result<Vec<u8>> {
	let mut body: Vec<u8> = vec![];
	while let Some(next) = req.frame().await {
		let frame = next?;
		if let Some(chunk) = frame.data_ref() {
			body.extend_from_slice(chunk);
		}
	}

	Ok(body)
}

pub async fn login_user(pool: DBPool, data: HashMap<String, String>) -> Res<(String, String), SkyError> {
	let mut ret: (String, String) = Default::default();
	let mut con = pool.get().await.unwrap();

	let login = &data.get("login").unwrap().trim().to_lowercase();
	let pass = data.get("password").unwrap().trim();

	let q = con
		.query_parse::<(String,)>(&query!("SELECT uuid FROM bitauth.users_uuid WHERE login = ?", login))
		.await;
	if q.is_err() { println!("{:?}", q.err()); return Ok(ret); };
	let (uuid,) = q.unwrap();

	let q = con
		.query_parse::<Users>(&query!("SELECT * FROM bitauth.users WHERE uuid = ?", uuid.clone()))
		.await;
	if q.is_err() {
		discord_wh_send(
			format!("Пизда тут, ну тип, да, какая-то ебень с uuid, чел потерялся...\n`{}`\n**{}**", uuid, login)
		).await;
		return Ok(ret);
	}

	// TODO: Say to user about trouble.

	let q = q.unwrap();
	if bcrypt::verify(pass, q.password.as_str()).unwrap() {
		let token = format!("{}", uuid_v4().as_hyphenated());
		let reftoken = format!("{}", uuid_v4().as_hyphenated());
		let time = time();
		let uuid: String = q.uuid;
		let login: String = q.login;

		let _ = con.query_parse::<()>(&query!(
			"INSERT INTO bitauth.tokens { uuid: ?, uid: ?, sid: ?, ref: ?, refend: ? }",
			token.clone(), uuid.clone(), "0", reftoken.clone(), time + REFRESH_LIFETIME
		)).await;
		let _ = con.query_parse::<()>(&query!(
			"UPDATE bitauth.users SET tokens += ? WHERE login = ?",
			token.clone(), login.clone()
		)).await;

		ret = (
			jwt_sign(pool.clone(), json!({
				"login": login.clone(),
				"uuid": uuid.clone(),
				"iat": time,
				"exp": time + TOKEN_LIFETIME
			})).await.unwrap(),
			jwt_sign(pool.clone(), json!({
				"uuid": token.clone(),
				"iat": time,
				"ref": reftoken.clone(),
				"exp": time + REFRESH_LIFETIME
			})).await.unwrap()
		);
	}

	Ok(ret)
}

pub async fn jwt_sign(pool: DBPool, data: Json) -> Result<String> {
	let mut con = pool.get().await.unwrap();

	let q = con
		.query_parse::<Sites>(&query!("SELECT * FROM bitauth.sites WHERE uuid = ?", "0"))
		.await;
	if q.is_err() { return Ok("".to_owned()) };
	let skey = q.unwrap().skey;

	let skey = EncodingKey::from_rsa_der(&skey);
	let header = Header::new(Algorithm::RS512);
	let token = jwt_encode(&header, &data, &skey)?;

	Ok(token)
}

pub fn rsa_gen() -> (Vec<u8>, Vec<u8>) {
	let mut rng = rand::thread_rng();
	let bits = 2048;
	let skey = RsaPrivateKey::new(&mut rng, bits).expect("RSA err");
	let pkey = RsaPublicKey::from(&skey);

	(
		skey.to_pkcs1_der().unwrap().as_bytes().to_vec(),
		pkey.to_pkcs1_der().unwrap().as_bytes().to_vec()
	)
}

pub async fn jwt_verify(pool: DBPool, token: &str) -> Result<TokenData<Json>> {
	let mut con = pool.get().await.unwrap();
	let mut ret = TokenData { header: Header::new(Algorithm::RS512), claims: json![{}] };

	let q = con
		.query_parse::<Sites>(&query!("SELECT * FROM bitauth.sites WHERE uuid = ?", "0"))
		.await;
	if q.is_err() { return Ok(ret) };
	let pkey = q.unwrap().pkey;

	let pkey = DecodingKey::from_rsa_der(&pkey);
	let token = jwt_decode::<Json>(&token, &pkey, &Validation::new(Algorithm::RS512));

	match token.is_ok() {
		true => ret = token.clone().unwrap(),
		_ => {}
	}

	Ok(ret)
}

pub async fn create_user(pool: DBPool, data: HashMap<String, String>) -> Res<bool, SkyError> {
	let mut ret = true;
	let mut con = pool.get().await.unwrap();

	let login = &data.get("login").unwrap().trim().to_lowercase();
	let email = &data.get("email").unwrap().trim().to_lowercase();
	let pass = data.get("password").unwrap().trim();
	let pass2 = data.get("password2").unwrap().trim();

	if pass != pass2 { ret = false };
	if pass.len() < 8 { ret = false };

	let q = con
		.query_parse::<(String,)>(&query!("SELECT uuid FROM bitauth.users_uuid WHERE login = ?", login))
		.await;
	if q.is_ok() { ret = false };

	if ret {
		let uuid = format!("{}", uuid_v4().as_hyphenated());
		let pass = bcrypt::hash(pass, 12).unwrap();

		let q = con.query_parse::<()>(&query!(
			r#"INSERT INTO bitauth.users {
				uuid: ?,
				login: ?,
				password: ?,
				email: ?,
				tokens: []
			}"#,
			uuid.clone(),
			login.clone(),
			pass,
			email,
		)).await;
		if q.is_err() { ret = false }

		let q = con.query_parse::<()>(&query!(
			r#"INSERT INTO bitauth.users_uuid {
				login: ?,
				uuid: ?
			}"#,
			login.clone(),
			uuid.clone()
		)).await;
		if q.is_err() { ret = false }
	}

	Ok(ret)
}

pub async fn get_user(pool: DBPool, login: String) -> Res<Users, SkyError> {
	let mut con = pool.get().await.unwrap();

	let q = con
		.query_parse::<Users>(&query!("SELECT * FROM bitauth.users WHERE login = ?", login))
		.await;

	Ok(q?)
}

pub fn get_cookies(headers: HeaderMap) -> HashMap<String, String> {
	let header = headers.get(hyper::header::COOKIE);
	let cookies = match header.is_none() {
		false => header.unwrap().to_str().unwrap(),
		_ => ""
	};

	double_split(cookies.to_owned(), ";", "=")
}

pub async fn authorize_user(pool: DBPool, token: String, session: String) {
	let mut con = pool.get().await.unwrap();

	let data: JsonMap<String, Json> = jwt_verify(pool.clone(), &token)
		.await
		.unwrap()
		.claims
		.as_object()
		.unwrap()
		.clone();

	let login = data.get("login").unwrap().as_str();
	let uuid = data.get("uuid").unwrap().as_str();

	let _ = con.query_parse::<()>(&query!(
		r#"INSERT INTO bitauth.v0 {
			session: ?,
			login: ?,
			uuid: ?,
			expire: ?
		}"#,
		session,
		login,
		uuid,
		time() + APIV0_LIFETIME
	)).await;
}
